pub mod config;
pub mod query;
#[cfg(test)]
mod tests;

use std::{
    fmt,
    fs::OpenOptions,
    io::{BufWriter, Write},
    str::FromStr,
    sync::{
        Arc, RwLock,
        mpsc::{self, Sender},
    },
    thread,
    time::SystemTime,
};

use config::LoggerConfig;
use query::LogQuery;
use tracing::{Event, Level, Subscriber};
use tracing_subscriber::{
    layer::{Context, Layer, SubscriberExt},
    registry::LookupSpan,
    util::SubscriberInitExt,
};

#[derive(Debug, Clone, PartialEq, Eq)]
enum LogEvent {
    Log(LogEntry),
    Shutdown,
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct LogEntry {
    timestamp: SystemTime,
    level: Level,
    message: String,
}

impl PartialOrd for LogEntry {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for LogEntry {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        self.timestamp
            .duration_since(SystemTime::UNIX_EPOCH)
            .unwrap()
            .cmp(
                &other
                    .timestamp
                    .duration_since(SystemTime::UNIX_EPOCH)
                    .unwrap(),
            )
    }
}

struct BufferLayer {
    log_entries: Arc<RwLock<Vec<LogEntry>>>,
    senders: Vec<Sender<LogEvent>>,
}

impl BufferLayer {
    pub fn new(log_entries: Arc<RwLock<Vec<LogEntry>>>, senders: Vec<Sender<LogEvent>>) -> Self {
        Self {
            log_entries,
            senders,
        }
    }
}

impl Drop for BufferLayer {
    fn drop(&mut self) {
        for tx in &self.senders {
            tx.send(LogEvent::Shutdown).unwrap();
        }
        self.senders.clear();
    }
}

impl<S> Layer<S> for BufferLayer
where
    S: Subscriber + for<'a> LookupSpan<'a>,
{
    fn on_event(&self, event: &Event<'_>, _ctx: Context<'_, S>) {
        let metadata = event.metadata();
        let level = *metadata.level();
        let timestamp = SystemTime::now();
        let mut message = String::new();
        let mut visitor = LogVisitor::new(&mut message);
        event.record(&mut visitor);

        let log_entry = LogEvent::Log(LogEntry {
            timestamp,
            level,
            message,
        });

        if let LogEvent::Log(ref entry) = log_entry {
            if let Ok(mut buf) = self.log_entries.write() {
                buf.push(entry.clone());
            }
        }

        for tx in &self.senders {
            let _ = tx.send(log_entry.clone());
        }
    }
}

struct LogVisitor<'msg> {
    message: &'msg mut String,
}

impl<'msg> LogVisitor<'msg> {
    fn new(message: &'msg mut String) -> Self {
        Self { message }
    }
}

impl tracing::field::Visit for LogVisitor<'_> {
    fn record_debug(&mut self, field: &tracing::field::Field, value: &dyn fmt::Debug) {
        use std::fmt::Write;
        if field.name() == "message" {
            write!(self.message, "{:?}", value).unwrap();
        } else {
            write!(self.message, "{}={:?} ", field.name(), value).unwrap();
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Default)]
pub enum LogLevel {
    Error,
    #[default]
    Info,
    Warn,
    Debug,
    Trace,
}

impl From<LogLevel> for tracing::Level {
    fn from(level: LogLevel) -> Self {
        match level {
            LogLevel::Error => Level::ERROR,
            LogLevel::Info => Level::INFO,
            LogLevel::Debug => Level::DEBUG,
            LogLevel::Trace => Level::TRACE,
            LogLevel::Warn => Level::WARN,
        }
    }
}

impl fmt::Display for LogLevel {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

impl FromStr for LogLevel {
    type Err = ();

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "trace" => Ok(Self::Trace),
            "debug" => Ok(Self::Debug),
            "info" => Ok(Self::Info),
            "error" => Ok(Self::Error),
            "warn" => Ok(Self::Warn),
            _ => Err(()),
        }
    }
}

pub struct Logger {
    subscriber_guard: Option<tracing::subscriber::DefaultGuard>,
    log_entries: Arc<RwLock<Vec<LogEntry>>>,
    _handles: Vec<thread::JoinHandle<()>>,
}

impl Logger {
    pub fn new(config: LoggerConfig) -> Self {
        let log_entries = Arc::new(RwLock::new(Vec::new()));
        let mut senders = Vec::new();
        let mut handles = Vec::new();

        if config.log_to_stdout {
            let (tx, rx) = mpsc::channel();
            senders.push(tx);
            let config_clone = config.clone();
            let handle = thread::spawn(move || {
                for msg in rx {
                    match msg {
                        LogEvent::Log(entry) => {
                            println!(
                                "{}",
                                format_entry(
                                    &entry,
                                    config_clone.stdout_color,
                                    config_clone.stdout_include_time
                                )
                            );
                        }
                        LogEvent::Shutdown => break,
                    }
                }
            });
            handles.push(handle);
        }

        if config.log_to_file {
            let (tx, rx) = mpsc::channel();
            senders.push(tx);
            let path = config.log_file_path.clone();
            let include_time = config.file_include_time;
            let handle = thread::spawn(move || {
                let file = OpenOptions::new()
                    .append(true)
                    .create(true)
                    .open(&path)
                    .expect("Failed to open log file");
                let mut writer = BufWriter::new(file);
                for msg in rx {
                    match msg {
                        LogEvent::Log(entry) => {
                            let line = format_entry(&entry, false, include_time);
                            writeln!(writer, "{}", line).expect("Failed to write to log file");
                            writer.flush().expect("Failed to flush log file");
                        }
                        LogEvent::Shutdown => break,
                    }
                }
            });
            handles.push(handle);
        }

        let buffer_layer = BufferLayer::new(log_entries.clone(), senders);
        let subscriber = tracing_subscriber::registry().with(buffer_layer);
        let guard = subscriber.set_default();

        Logger {
            subscriber_guard: Some(guard),
            log_entries,
            _handles: handles,
        }
    }

    pub fn get_logs(&self, query: LogQuery) -> Vec<LogEntry> {
        let guard = self.log_entries.read().unwrap();
        match query {
            LogQuery::All => guard.clone(),
            LogQuery::From(t) => guard.iter().filter(|e| e.timestamp >= t).cloned().collect(),
        }
    }
}

impl Drop for Logger {
    fn drop(&mut self) {
        let _ = self.subscriber_guard.take();
        let handles = std::mem::take(&mut self._handles);
        for handle in handles {
            handle.join().unwrap_or_else(|e| {
                eprintln!("Logger thread panicked: {:?}", e);
            });
        }
    }
}

fn format_entry(entry: &LogEntry, use_color: bool, _: bool) -> String {
    let lvl = if use_color {
        match entry.level {
            Level::ERROR => "\x1b[31mERROR\x1b[0m",
            Level::WARN => "\x1b[33mWARN\x1b[0m",
            Level::INFO => "\x1b[32mINFO\x1b[0m",
            Level::DEBUG => "\x1b[36mDEBUG\x1b[0m",
            Level::TRACE => "\x1b[34mTRACE\x1b[0m",
        }
    } else {
        entry.level.as_str()
    };

    format!("{} {}", lvl, entry.message)
}
